/**************************************************************/
/*                                                            */
/* Program Name:  ACORDASC.C                                  */
/*                                                            */
/*  Program to process an ACORD DATADICT.ASC file and         */
/*  generate meta data files.                                 */
/*                                                            */
/*  Written by Jim MacNair - September, 2000                  */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  Current Version 1.00                                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  This program handles ACORD ASC file with segment          */
/*   definitions.  It creates the meta data files for each    */
/*   segment.                                                 */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/* This program takes an ACORD ASCII file with definitions    */
/* of the various ACORD segments and processes the            */
/* definitions into meta data repository files.               */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_LINE      121
#define MAX_LINE_LEN  1024
#define MAX_VARIABLES 1024
#define MAX_MSG_DATA_LEN 65535
#define ACORDSTDLEN 10
#define DELETED_NAME  "ZZDEL"
#define RESERVED_NAME "ZZZZZ"

#define NAME_OPTION 2

#include "mtdstruc.h"
#include "ACORD.h"

typedef struct {
   int  VarLevel;
   int  VarOffset;
   int  VarLen;
   int  VarParent;
   int  VarChild;
   int  VarLastChild;
   int  VarNextLevel;
   int  VarPrevLevel;
   int  VarOccur;
   char	VarFlags;
   char	VarFlags2;
   char VarDecPoint;
   char VarType;
   char VarUsage;
   int VarNamePtr;
   int VarUnamePtr;
} VARSTRUCT;

typedef struct {
	char	seg[5];
	char	nam[6];
	char	typ;
	int		occ;
	int		first;
	int		last;
	char	newname[40];
} REPSTRUCT;

/**************************************************************/
/* Variable table size definitions                            */
/**************************************************************/

#define MAX_VAR_COUNT  4096

static char HEXTABLE[] = "0123456789ABCDEF";

static char copyright[] = "(C) Copyright IBM Corp, 2000";
static char Version[]=\
"@(#)ACORDASC V1.00 - ACORD ASC file processing program  - Jim MacNair ("__DATE__")";

static char nameArea[65536];
static int nameLen=0;
static wchar_t UnameArea[65536];
static int UnameLen=0;
static char defaultnames[65536];
static REPSTRUCT rep[1024];

/*****************************************************************/
/*                                                               */
/* Global Debug Variable - If debugging information is desired,  */
/*  set this variable to 1 and recompile.                        */
/*                                                               */
/*****************************************************************/

DEBUG=0;

/**************************************************************/
/*                                                            */
/* This routine skips blanks in a string                      */
/*                                                            */
/**************************************************************/

char * skipBlanks(char *str)

{
   while (*str == ' ')
   {
      str++;
   } /* endwhile */

   return str;
}

/**************************************************************/
/*                                                            */
/* This routine finds the next blank in a string              */
/*                                                            */
/**************************************************************/

char * findBlank(char *str)

{
   while ((str[0] != ' ') && (str[0] != 0))
   {
      str++;
   } /* endwhile */

   return str;
}

/**************************************************************/
/*                                                            */
/* This routine will change any occurances of the specified   */
/* character into a blank.                                    */
/*                                                            */
/**************************************************************/

void changeChar(char *inStr, char oldch, char newch)

{
	char	*tempptr;

	tempptr = strchr(inStr, oldch);
	while (tempptr != NULL)
	{
		tempptr[0] = newch;
		tempptr = strchr(tempptr,oldch);
	}
}

/**************************************************************/
/*                                                            */
/* This routine writes a line to the message file             */
/* that is produced with the new program.                     */
/*                                                            */
/**************************************************************/

void writeErrorLine(int count, char *msg, FILE *msgfile)

{
   /* print the message */
   fprintf(msgfile,"line %.4d - %s\n",count + 1,msg);
}

/*******************************************************************/
/*                                                                 */
/*  Subroutine to convert the data to a hexidecimal                */
/*   representation.                                               */
/*                                                                 */
/*******************************************************************/

void buildHex(char *dataText, int len, const char *msgdata)

{
	int	i;
	int	j=0;
	char ch;

	for (i=0; i < len; i++)
	{
		ch = (unsigned char) msgdata[i] >> 4;
		ch = HEXTABLE[ch];
		//data2hex(&ch);
		dataText[j++] = ch;

		ch = (unsigned char) msgdata[i] & 0x0F;
		ch = HEXTABLE[ch];
		//data2hex(&ch);
		dataText[j++] = ch;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to write a segment name mapping file, based on the */
/*  data in the defaultnames data.                            */
/*                                                            */
/**************************************************************/

void writeSegnameFile(FILE* segfile)

{
	char*	ptr;

	ptr=defaultnames;
	while (ptr[0] != 0)
	{
		// write an entry to the segment name file
		fprintf(segfile, "%4.4s%s\n", ptr, ptr + 5);

		ptr += strlen(ptr) + 1;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to load the repeated data items file.              */
/*                                                            */
/**************************************************************/

void loadRepeat()

{
	int		i;
	FILE*	repfile;
	char	workarea[32];
	char	reprec[64];

	// initialize the memory to zeros
	memset(rep, 0, sizeof(rep));
	memset(reprec, 0, sizeof(reprec));

	// try to open the repeat.dat file
	repfile = fopen("REPEAT.DAT", "r");

	// did the open work?
	if (repfile != NULL)
	{
		// initialize counter
		i = 0;

		// read the file
		while (fgets(reprec, sizeof(reprec), repfile) != NULL)
		{
			// eliminate the NL character on the end of the input buffer
			if (reprec[strlen(reprec) - 1] == '\n')
			{
				reprec[strlen(reprec) - 1] = 0;
			}

			// build the next table entry
			memcpy(rep[i].seg, reprec, 4);
			memcpy(rep[i].nam, reprec + 4, 5);
			rep[i].typ = reprec[9];
			memset(workarea, 0, sizeof(workarea));
			memcpy(workarea, reprec + 10, 4);
			rep[i].occ = atoi(workarea);
			memset(workarea, 0, sizeof(workarea));
			memcpy(workarea, reprec + 14, 4);
			rep[i].first = atoi(workarea);
			memset(workarea, 0, sizeof(workarea));
			memcpy(workarea, reprec + 18, 4);
			rep[i].last = atoi(workarea);
			strcpy(rep[i].newname, reprec + 22);

			// clear the memory before we read the next record
			memset(reprec, 0, sizeof(reprec));

			// go on to the next entry
			i++;
		}

		// close the file
		fclose(repfile);
	}
}

REPSTRUCT* checkRepeat(DDE_REC* dde_ptr, char * repType, int* occurs)

{
	int		i=0;
	int		iseq;
	REPSTRUCT*	ptr=NULL;
	char	seqno[3];

	/* Initialize the return values */
	(*occurs) = 1;
	repType[0] = 0;

	/* get the sequence number of this item as an integer */
	memset(seqno, 0, sizeof(seqno));
	memcpy(seqno, dde_ptr->seq_no, sizeof(dde_ptr->seq_no));
	iseq = atoi(seqno) + 1;

	/* point to the beginning of the repeated items table */
	ptr = rep;
	while ((ptr->nam[0] != 0) && 
		   ((memcmp(dde_ptr->reference_name, ptr->nam, sizeof(dde_ptr->reference_name)) != 0) ||
		   (memcmp(dde_ptr->group_desig, ptr->seg, sizeof(dde_ptr->group_desig)) != 0) ||
		   (iseq != ptr->first)))
	{
		ptr++;
	}

	/* did we find the element? */
	if (ptr->nam[0] == 0)
	{
		ptr = NULL;
	}
	else
	{
		/* structure */
		(*occurs) = ptr->occ;
		repType[0] = ptr->typ;
	}

	return ptr;
}

/**************************************************************/
/*                                                            */
/* Routine to build an initialization string.                 */
/*                                                            */
/**************************************************************/

void initialize(char * outBuf, VARSTRUCT* fvar, int firstVar, int lastVar)

{
	int	i;
	int j;
	int k;
	int	occur;
	int	length;
	int	var1;
	int	var2;
	int addlOfs;
	unsigned char value=0;
	unsigned char * bufOfs;

	i=firstVar;
	while(i < lastVar)
	{
		// calculate the number of occurences of this variable
		// this will return a 1 if there is no occurs count
		occur = 1;
		if (fvar[i].VarOccur > 1)
		{
			occur = fvar[i].VarOccur;
		}

		// check if we have an elementary variable
		if (fvar[i].VarChild == 0)
		{
			// initialize this variable,
			// handling any occurs clauses
			bufOfs = outBuf + fvar[i].VarOffset;

			for (j=0; j < occur; j++)
			{
				// Get the data into a work area
				length = fvar[i].VarLen;
				
				// figure out the type of the data
				// check for numeric rather than character data
				if ((fvar[i].VarFlags & FLAGS_NUMERIC) > 0)
				{
					memset(bufOfs, '0', length);
				}
				else
				{
					memset(bufOfs, ' ', length);
				}

				bufOfs += length;
			}
		}
		else
		{
			// must be higher level group item
			// only need to initialize if subject of occurs
			if (occur > 1)
			{
				// initialize this variable,
				// handling any occurs clauses
				addlOfs = 0;
				var1 = fvar[i].VarChild;
				var2 = fvar[i].VarLastChild + 1;
				length = fvar[i].VarLen;

				k = 0;
				while (k < occur)
				{
					// We will recursively enter the initialize routine
					initialize(outBuf + addlOfs, fvar, var1, var2);
					addlOfs += length;
					k++;
				}

				// We already processed the entire group - no need to do it again
				i = var2 - 1;
			}
		}

		i++;
	}
}

void replaceName(char * longName, const char * segName)

{
	char* ptr;

	ptr = defaultnames;
	while ((ptr[0] != 0) && (memcmp(segName, ptr, 4) != 0))
	{
		ptr += strlen(ptr) + 1;
	}

	if (ptr[0] != 0)
	{
		strcpy(longName, ptr + 5);
	}
}

/*******************************************************************/
/*                                                                 */
/* Routine to process a variable definition.                       */
/*                                                                 */
/*******************************************************************/

void writeElement(VARSTRUCT* fvar, FILE* mtdfile, FILE* msgfile)

{
	int				items;
	VAR_FILE_REC	varFileData;

	varFileData.Level = fvar->VarLevel;
	varFileData.Offset = fvar->VarOffset;
	varFileData.Length = fvar->VarLen;
	varFileData.Parent = fvar->VarParent;
	varFileData.FirstChild = fvar->VarChild;
	varFileData.LastChild = fvar->VarLastChild;
	varFileData.NextSibling = fvar->VarNextLevel;
	varFileData.PrevSibling = fvar->VarPrevLevel;
	varFileData.Occur = fvar->VarOccur;
	varFileData.OccurMax = 0;
	varFileData.OccurDepend = 0;
	varFileData.Usage = fvar->VarUsage;
	varFileData.Flags  = fvar->VarFlags;
	varFileData.Flags2 = fvar->VarFlags2;
	varFileData.decimalPoint = fvar->VarDecPoint;
	varFileData.Reserved  = 0;
	varFileData.Reserved2 = 0;
	varFileData.Reserved3 = 0;
	varFileData.Nameptr = fvar->VarNamePtr;
	varFileData.Unameptr = fvar->VarUnamePtr;

	/* write the data to the file */
	items = fwrite(&varFileData, sizeof(varFileData), 1, mtdfile);
	fflush(mtdfile);

	/* tell what we are doing in the message file */
	fprintf(msgfile, "%2d %4d %4d %s\n", fvar->VarLevel, fvar->VarOffset, fvar->VarLen, nameArea + fvar->VarNamePtr);
}

/**************************************************************/
/*                                                            */
/* Routine to write the variable table to a metadata file.    */
/*                                                            */
/**************************************************************/

void writeVars(VARSTRUCT* fvar, 
			   char * segName, 
			   char * segVersion, 
			   int maxvar, 
			   FILE* msgfile)

{
	int		i;
	char	initStr[1000];		/* data area for initialization string */
	char	charlen[10];
	META_FILE_HEADER hdr;
    FILE	*mtdfile;			/* output mtd file                    */
	char	mtdFileName[255];	/* Name of the output meta data file  */

	// Creat the first part of the file name, using the group identifier
	strcpy(mtdFileName, segName);
	if (segVersion[0] != ' ')
	{
		strcat(mtdFileName, segVersion);
	}

	// Finish the meta data file name by adding an extension
	strcat(mtdFileName, ".mtd");

	if ((mtdfile = fopen(mtdFileName, "wb")) == NULL)
	{
		fprintf(msgfile, "Unable to open meta data file - %s\n", mtdFileName);
	}
	else
	{
		fprintf(msgfile, "Writing meta data file %s\n", mtdFileName);

		/* write out the file header first */
		memset(&hdr, 0, sizeof(hdr));
		hdr.HdrLength  = sizeof(hdr);
		hdr.Version = 0;
		hdr.Seglen = fvar[0].VarLen;
		hdr.Maxvar = maxvar;
		hdr.Names = sizeof(hdr) + (maxvar * sizeof(VAR_FILE_REC));
		hdr.Unames = hdr.Names + nameLen + 1;
		hdr.InitialValue = hdr.Unames + ((UnameLen + 1) * sizeof(wchar_t));
		fwrite(&hdr, sizeof(hdr), 1, mtdfile);

		/* write out the elements in this meta data file */
		i = 0;
		while (i < maxvar)
		{
			writeElement(&fvar[i], mtdfile, msgfile);
			i++;
		}

		/* write out the names and Unicode names areas */
		fwrite(nameArea, nameLen + 1, 1, mtdfile);
		fwrite(UnameArea, UnameLen + 1, sizeof(UnameArea[0]), mtdfile);

		/* Finally, build and write out the initialization string */
		memset(initStr, 0, sizeof(initStr));
		initialize(initStr, &fvar[0], 1, maxvar);

		// fill in the first part of the header
		memcpy(initStr, segName, 4);
		itoa(fvar[0].VarLen, charlen, 10);
		memcpy(initStr + 7 - strlen(charlen), charlen, strlen(charlen));

		// see if there is a version number for this group
		if ((strlen(segVersion) == 1) && (segVersion[0] > '0'))
		{
			initStr[8] = segVersion[0];
		}

		// write the initialization string out to the metadata file
		fwrite(initStr, fvar[0].VarLen, 1, mtdfile);

		/* First, close the current file */
		fclose(mtdfile);
	}
}

/**************************************************************/
/*                                                            */
/* Routine to save the variable type.                         */
/*                                                            */
/**************************************************************/

void saveType(DDE_REC* dde_ptr, VARSTRUCT* fvar, int maxvar)

{
	char	workArea[3];

	/* capture the data type in a work area */
	memcpy(workArea, dde_ptr->data_type, sizeof(dde_ptr->data_type));
	workArea[sizeof(dde_ptr->data_type)] = 0;

	switch (workArea[0])
	{
	case 'A':
		{
			fvar[maxvar].VarType = USE_DISPLAY;
			if (workArea[1] == ' ')
			{
				fvar[maxvar].VarFlags |= FLAGS_ALPHABETIC;
			}

			break;
		}
	case 'N':
		{
			fvar[maxvar].VarType = USE_DISPLAY;
			fvar[maxvar].VarFlags |= FLAGS_NUMERIC;
			if ((workArea[1] >= '0') && (workArea[1] <= '9'))
			{
				fvar[maxvar].VarDecPoint = atoi(workArea + 1);
			}

			break;
		}
	case 'S':
		{
			fvar[maxvar].VarType = USE_DISPLAY;
			fvar[maxvar].VarFlags |= (FLAGS_NUMERIC | FLAGS_SIGNED);
			if ((workArea[1] >= '0') && (workArea[1] <= '9'))
			{
				fvar[maxvar].VarDecPoint = atoi(workArea + 1);
			}

			break;
		}
	case 'T':
		{
			fvar[maxvar].VarType = USE_DISPLAY;
			fvar[maxvar].VarFlags |= FLAGS_TEXT;

			break;
		}
	}

	if ((memcmp(dde_ptr->reference_name, DELETED_NAME, sizeof(DELETED_NAME) - 1) == 0) ||
		(memcmp(dde_ptr->reference_name, RESERVED_NAME, sizeof(RESERVED_NAME) - 1) == 0))
	{
		/* Set the filler flag */
		fvar[maxvar].VarFlags2 |= FLAGS_FILLER;
	}

	/* check for Yes/No data type */
	if (memcmp(dde_ptr->field_class, "Y/N", 3) == 0)
	{
		fvar[maxvar].VarFlags |= FLAGS_YES_NO;
	}

	/* check for date type */
	if (memcmp(dde_ptr->field_class, "DT", 2) == 0)
	{
		fvar[maxvar].VarFlags |= FLAGS_DATE;
	}

	/* check for phone number data type */
	if (memcmp(dde_ptr->field_class, "PH", 2) == 0)
	{
		fvar[maxvar].VarFlags |= FLAGS_PHONE;
	}

	/* check for Percentage type */
	if (memcmp(dde_ptr->field_class, "PCT", 3) == 0)
	{
	}

	/* check for Time type */
	if (memcmp(dde_ptr->field_class, "TI", 2) == 0)
	{
		fvar[maxvar].VarFlags |= FLAGS_TIME;
	}

			/* check for mod type */
	if (memcmp(dde_ptr->field_class, "MOD", 3) == 0)
	{
	}

	/* check for DC type */
	if (memcmp(dde_ptr->field_class, "DC", 2) == 0)
	{
	}

	/* check for IC type */
	if (memcmp(dde_ptr->field_class, "IC", 2) == 0)
	{
	}

	/* check for WD type */
	if (memcmp(dde_ptr->field_class, "WD", 2) == 0)
	{
	}

	/* check for special data types */
	if (memcmp(dde_ptr->data_type, "   ", 3) != 0)
	{
	}
}

/**************************************************************/
/*                                                            */
/* Routine to save variable name in name tables.              */
/*                                                            */
/**************************************************************/

void saveName(const char * name)

{
	int		size;
	wchar_t	UtempName[256];

	// first, append this name to the name table
	strcpy(nameArea + nameLen, name);
	nameLen += strlen(name) + 1;

	// convert the long name to Unicode, for efficiency at runtime
	size = mbstowcs(UtempName, name, sizeof(UtempName) - 1);

	// append it to the Unicode name table
	wcscpy(UnameArea + UnameLen, UtempName);
	UnameLen += wcslen(UtempName) + 1;
}

/**************************************************************/
/*                                                            */
/* Routine to create a segment name.                          */
/*                                                            */
/* The name will be built from the description field.  This   */
/* field can be a maximum of 60 characters.  Since very long  */
/* names can be very cumbersome, the name generated is        */
/* limited to less than 30 characters.  In some cases, a      */
/* duplicate name would be generated.  The cases where this   */
/* happens have been identified and alternate names have been */
/* hard coded into this program.  Only complete words are     */
/* used to build the application name.  Connective words are  */
/* dropped (of and or).                                       */
/*                                                            */
/**************************************************************/

void buildSegNameEntry(FILE* segfile, FILE* msgfile, DDE_REC* dde_ptr, 
					   const char* segName, char * longName)

{
	char	tempName[128];
	wchar_t	UtempName[256];
	char*	ptrDesc;
	char*	ptrName;
	int		upcase=0;
	int		i;
	int		j;
	int		namedone=0;
	int		count=0;
	int		size;

	// Initialize the work area for the variable name
	memset(tempName, 0, sizeof(tempName));

	// try to find a default name for this group
	replaceName(tempName, segName);

	if (strlen(tempName) == 0)
	{
		// Point to the input and output areas
		ptrDesc = dde_ptr->element_name;
		ptrName = tempName;

		i=0;
		while ((i < sizeof(dde_ptr->element_name)) && (namedone == 0))
		{
			// Limit names to less than 30 characters by stopping if
			// the next word in the description would put us past the limit
			// Are we at the beginning of a word?
			if (ptrDesc[0] == ' ')
			{
				// Yes, beginning of another word
				// figureout how long it is
				j = 1;
				while (ptrDesc[j] != ' ')
				{
					j++;
				}

				if ((count + j) > 29)
				{
					fprintf(msgfile, "Name truncated because it would exceed 30 characters\n");
					break;
				}

				upcase = 0;
			}

			if (((ptrDesc[0] >= 'a') && (ptrDesc[0] <= 'z')) ||
				((ptrDesc[0] >= 'A') && (ptrDesc[0] <= 'Z')))
			{
				if ((memcmp(ptrDesc, "of ", 3) == 0) || (memcmp(ptrDesc, "or ", 3) == 0) ||
					(memcmp(ptrDesc, "at ", 3) == 0))
				{
					// skip of or at
					ptrDesc += 2;
				}
				else
				{
					// check if this is not the first character in a word
					// and the letter is upper case
					if ((upcase == 1) && (ptrDesc[0] >= 'A') && (ptrDesc[0] <= 'Z'))
					{
						ptrDesc[0] = ptrDesc[0] + 'a' - 'A';
					}

					ptrName++[0] = ptrDesc[0];
					count++;
				}

				upcase = 1;
			}

			ptrDesc++;
			i++;
		}

		// Check if this is an early version of the dictionary file
		// If so, the segment name will be header and we should change it
		// to be the name of the segment with a leading underscore
		if (strcmp(tempName, "Header") == 0)
		{
			tempName[0] = '_';
			tempName[1] = 0;
			strcat(tempName, segName);
		}

		// Eliminate the word group or groups from the end of all the descriptions
		if ((strcmp(ptrName - 5, "Group") == 0) || (strcmp(ptrName - 5, "GROUP") == 0))
		{
			ptrName -= 5;
			ptrName[0] = 0;
			count -= 5;
		}

		if ((strcmp(ptrName - 6, "Groups") == 0) || (strcmp(ptrName - 6, "GROUPS") == 0))
		{
			ptrName -= 6;
			ptrName[0] = 0;
			count -= 6;
		}

		printf("Generated name for segment %s name %s\n", segName, tempName);
	}

	// save the name in the character name area
	saveName(tempName);

	// convert the long name to Unicode, for efficiency at runtime
	size = mbstowcs(UtempName, tempName, sizeof(UtempName) - 1);

	fprintf(msgfile, "Name (%S) assigned to segment (%s) chars (%d) size (%d)\n", 
			UtempName, segName, count, size);

	strcpy(longName, tempName);
}

/**************************************************************/
/*                                                            */
/* Routine to build a variable name from the dictionary.      */
/*                                                            */
/**************************************************************/

void buildName(DDE_REC* dde_ptr, char* tempName, int maxvar)

{
	int	i;
	int	j;
	int k;
	int dupfound;
	char *varPtr;			/* Working pointer to variable name   */

	if ((memcmp(dde_ptr->reference_name, DELETED_NAME, sizeof(DELETED_NAME) - 1) == 0) ||
		(memcmp(dde_ptr->reference_name, RESERVED_NAME, sizeof(RESERVED_NAME) - 1) == 0))
	{
		/* Use the reference name instead of the long name */
		memcpy(tempName, dde_ptr->reference_name, sizeof(dde_ptr->reference_name));
		i = strlen(tempName);
	}
	else
	{
		if (NAME_OPTION == 1)
		{
			memcpy(tempName, dde_ptr->element_name, sizeof(dde_ptr->element_name));

			/* remove any invalid characters by changing them to legal characters */
			changeChar(tempName, '/', '_');
			changeChar(tempName, '-', '_');
			changeChar(tempName, '\'', ' ');
			changeChar(tempName, '#', ' ');
			changeChar(tempName, '[', ' ');
			changeChar(tempName, ']', ' ');
			changeChar(tempName, '(', ' ');
			changeChar(tempName, ')', ' ');

			i = 0;
			while ((i < 27) && (tempName[i] > ' ') && (tempName[i] != '\''))
			{
				i++;
			}

			/* create a proper string */
			tempName[i] = 0;

			/* find the last word in the element name */
			j = sizeof(dde_ptr->element_name) - 1;
			while ((tempName[j] == ' ') || (tempName[j] == ')'))
			{
				tempName[j] = 0;
				j--;
			}

			while ((tempName[j] > ' ') && (tempName[j] != '('))
			{
				j--;
			}

			/* did we find a last word? */
			if ((i <= j) && (tempName[i + 1] > ' '))
			{
				tempName[i++] = '_';
				if (i <= j)
				{
					tempName[i] = 0;
					while ((i < 27) && (tempName[j + 1] > ' '))
					{
						tempName[i++] = tempName[j + 1];
						tempName[i] = 0;
						j++;
					}
				}
				else
				{
					/* Handle case of just two names separated by a blank */
					i = strlen(tempName);
				}
			}

			/* make sure we have room for the rest of the name */
			if (i > MAX_NAME_LEN - 4)
			{
				i = MAX_NAME_LEN - 4;
			}

			tempName[i++] = '_';
			memcpy(tempName + i, dde_ptr->seq_no, sizeof(dde_ptr->seq_no));
			tempName[i + sizeof(dde_ptr->seq_no)] = 0;
		}

		if (NAME_OPTION == 2)
		{
			/* Create a name from the reference name and the sequence number */
			/* get the reference name */
			i = sizeof(dde_ptr->reference_name) - 1;
			memcpy(tempName, dde_ptr->reference_name, sizeof(dde_ptr->reference_name));

			/* remove any trailing blanks */
			while ((i > 0) && (tempName[i] == ' '))
			{
				tempName[i--] = 0;
			}

			i++;
			tempName[i] = 0;

			/* check if this name occurs earlier */
			k = 0;
			dupfound=0;
			varPtr = nameArea;
			while (k < maxvar)
			{
				if (memcmp(tempName, varPtr, strlen(tempName)) == 0)
				{
					dupfound = 1;
				}

				varPtr += strlen(varPtr) + 1;
				k++;
			}

			/* Reference name is not unique - must append sequence number */
			if (dupfound)
			{
				/* insert a separator */
				tempName[i++] = '_';

				/* get the sequence number */
				memcpy(tempName + i, dde_ptr->seq_no, sizeof(dde_ptr->seq_no));
				i += sizeof(dde_ptr->seq_no);

				/* terminate the string */
				tempName[i] = 0;
			}
		}
	}
}

/**************************************************************/
/*                                                            */
/* Routine to create a standards mapping entry.               */
/*                                                            */
/**************************************************************/

void buildDefaultEntry(FILE* stdfile, 
					   FILE* msgfile,
					   FILE* deffile,
					   DDE_REC* dde_ptr,
					   char* segName,
					   char* segVersion,
					   char* currStdfile)

{
	char stdArea[32];       /* Work area for standard records     */
	char *currPtr;			/* Work pointer                       */
	int  dupfound;			/* Indicator if duplicate found       */

	memset(stdArea, 0, sizeof(stdArea));

	/* get any file name suffix characters */
	if ((dde_ptr->group_ver_no > ' ') && (dde_ptr->group_ver_no != 'b'))
	{
		segVersion[0] = dde_ptr->group_ver_no;
	}
	else
	{
		if (dde_ptr->ver_no[0] > '0')
		{
			memcpy(segVersion, dde_ptr->ver_no, 2);
		}
		else
		{
			segVersion[0] = ' ';
		}
	}

	// Create a standards file entry
	memset(stdArea, 0, sizeof(stdArea));
	memcpy(stdArea, dde_ptr->std_no, sizeof(dde_ptr->std_no));
	memcpy(stdArea + strlen(stdArea), dde_ptr->ver_no, sizeof(dde_ptr->ver_no));
	strcat(stdArea, segName);
	strcat(stdArea, segVersion);

	// check if there is a segment version
	dupfound = 0;
	currPtr = currStdfile;
	if (segVersion[1] == 0)
	{
		// now, check if it is already in the standards file data
		while ((currPtr[0] != 0) && (memcmp(currPtr, stdArea, strlen(stdArea)) != 0))
		{
			if (memcmp(currPtr, stdArea, strlen(stdArea) - 1) == 0)
			{
				if (segVersion[0] != currPtr[9])
				{
					fprintf(msgfile, "Different group version in current stdfile entry (%c) is different than in this file (%s) for %s\n",
							currPtr[9], segVersion, stdArea);
					printf("Different group version in current stdfile entry (%c) is different than in this file (%s) for %s\n",
							currPtr[9], segVersion, stdArea);
				}
			}

			// go on to the next entry
			currPtr += ACORDSTDLEN;
		}

		if (currPtr[0] != 0)
		{
			dupfound = 1;
		}
	}

	// Is it already in the file and do we want to write it?
	if ((dupfound == 0) && (segVersion[1] == 0))
	{
		// Lastly, check if the standard number is blank
		if (stdArea[0] > ' ')
		{
			// Add this entry if it doesn't already exist
			fprintf(stdfile, "%s", stdArea);

			// Indicate that we inserted a standard file entry
			fprintf(msgfile, "Standard number inserted (%s)\n", stdArea);
		}
		else
		{
			// Indicate that the standard number is blank
			fprintf(msgfile, "Standard number is blank (%s)\n", stdArea);
		}
	}
	else
	{
		if (segVersion[1] != 0)
		{
			// write a message to indicate the duplicate
			fprintf(msgfile, "No segment version - entry not added (%s)\n", stdArea);
		}
		else
		{
			// Indicate tha there is no segment version
			fprintf(msgfile, "Duplicate entry exists - not added (%s)\n", stdArea);
		}
	}

	// Create a defaults file entry
	strcpy(stdArea, segName);
	strcat(stdArea, segVersion);
	if (segVersion[1] == 0)
	{
		// Make all entries the same length
		strcat(stdArea, " ");
	}

	fprintf(deffile, "%s", stdArea);
}

/**************************************************************/
/*                                                            */
/* Routine to build a variable table entry.                   */
/*                                                            */
/**************************************************************/

void buildHeaderEntry(VARSTRUCT* fvar, 
					  int level,
					  int offset,
					  int length,
					  int parent,
					  int first,
					  int last,
					  int next,
					  int prev)

{
	fvar->VarLevel     = level;
	fvar->VarOffset    = offset;
	fvar->VarLen       = length;
	fvar->VarOccur     = 1;
	fvar->VarParent    = parent;
	fvar->VarChild     = first;
	fvar->VarLastChild = last;
	fvar->VarNextLevel = next;
	fvar->VarPrevLevel = prev;
	fvar->VarUsage     = USE_DISPLAY;
	fvar->VarNamePtr   = nameLen;
	fvar->VarUnamePtr  = UnameLen; 
}

/**************************************************************/
/*                                                            */
/* Routine to build a header area.                            */
/*                                                            */
/**************************************************************/

int buildHeader(VARSTRUCT* fvar, int length) 

{
	int maxvar=1;

	/* first, add the group variable and the fields */
	/* in the first ten bytes.                      */
	buildHeaderEntry(&fvar[maxvar], 5, 0, length, 0, maxvar + 1, maxvar + 5, maxvar + 6, 0);
	saveName("Header");
	maxvar++;

	buildHeaderEntry(&fvar[maxvar], 10, 0, 4, 1, 0, 0, maxvar + 1, 0);
	saveName("Identifier");
	maxvar++;

	buildHeaderEntry(&fvar[maxvar], 10, 4, 3, 1, 0, 0, maxvar + 1, maxvar - 1);
	fvar[maxvar].VarFlags |= FLAGS_NUMERIC;
	saveName("Length");
	maxvar++;

	buildHeaderEntry(&fvar[maxvar], 10, 7, 1, 1, 0, 0, maxvar + 1, maxvar - 1);
	saveName("Format");
	maxvar++;

	buildHeaderEntry(&fvar[maxvar], 10, 8, 1, 1, 0, 0, maxvar + 1, maxvar - 1);
	saveName("Version");
	maxvar++;

	buildHeaderEntry(&fvar[maxvar], 10, 9, 1, 1, 0, 0, 0, maxvar - 1);
	fvar[maxvar].VarFlags2 |= FLAGS_FILLER;
	saveName("Reserved");
	maxvar++;

	if (length > 10)
	{
		fvar[maxvar - 1].VarNextLevel = maxvar;

		buildHeaderEntry(&fvar[maxvar], 10, 10, 2, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("ProcLevel");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 12, 4, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("Iteration");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 16, 4, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("Ref_Identifier");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 20, 2, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("Ref_ProcLevel");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 22, 4, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("Ref_Iteration");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 26, 3, 1, 0, 0, maxvar + 1, maxvar - 1);
		saveName("ActionFlag");
		maxvar++;

		buildHeaderEntry(&fvar[maxvar], 10, 29, 1, 1, 0, 0, 0, maxvar - 1);
		fvar[maxvar].VarFlags2 |= FLAGS_FILLER;
		saveName("Spare");
		maxvar++;

		fvar[1].VarLastChild = maxvar - 1;
	}

	/* set the length of the root element */
	fvar[0].VarLen = length;

	return maxvar;
}

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void writeFileName(char *fileName, FILE *msgfile)

{
	/* write the file name to the message file */
	fprintf(msgfile, "Processing file %s\n", fileName);
	fflush(msgfile);
}

/**************************************************************/
/*                                                            */
/* Routine to write the line counter to the message file.     */
/*                                                            */
/**************************************************************/

void writeLinesRead(int inCount, FILE *msgfile, char *filename)

{
	/* write the line count to the message file */
   fprintf(msgfile,"Input lines read %5d  file %s\n",inCount,filename);
}

/**************************************************************/
/*                                                            */
/* Routine to display the command syntax.                     */
/*                                                            */
/**************************************************************/

void printHelp(char *cmd)
{
   printf("\nACORDASC V1.00\n");
   printf("\nCommand usage is:\n\n");
   printf("  %s ACORD-ASC-file-name\n\n", cmd);
   printf("A report file is printed with with the name of\n");
   printf("the ACORD ASC file and an extension of msg\n");
}

/**************************************************************/
/*                                                            */
/* Routine to create the meta data files from the ASC data    */
/*  dictionary file.                                          */
/*                                                            */
/**************************************************************/

int main(int argc, char *argv[])

{
	int  level;             /* Level number of current line       */
	int  length;            /* Length of elementary item          */
	int  curoffset;         /* Offset of elementary item          */
	int  maxvar=0;          /* Number of fields found             */
	int  i;                 /* Work variable                      */
	int  j;                 /* Work variable                      */
	int  occurs;			/* Number of occurences of variable   */
	int  currlen;			/* work variable to calculate length  */
	int  skipCount;			/* Number of entries to skip          */
	int  saveParent;		/* Current parent variable            */
	int  parent;            /* Work variable for parent chain     */
	int  defdone;           /* Work variable for parent chain     */
	int	RC=0;				/* Return code from program           */
	int	lastAtLevel[50];	/* Last element at this level         */
	DDE_REC* dde_ptr;		/* pointer to data definition record  */
	REPSTRUCT* repPtr=NULL;	/* Pointer to a repeating data struc  */
	char *fileExt;          /* file extension                     */
	char *currStdfile;		/* Pointer to previous stdfile data   */
	char workArea[MAX_LINE_LEN + 1];
	char repType;			/* Group or single repeating struct   */
	char currType[5];		/* Current group id                   */
	char dataArea[MAX_MSG_DATA_LEN + 1];
	char segVersion[3];     /* Segment version number             */
	char segName[5];		/* Work area to get segment id        */
	char tempName[128];     /* Work area for variable name        */
	char msgFileName[255];  /* name of message file               */
	char dataFileName[255]; /* name of input ACORD ASC file       */
	char tempArea[65536];
	char buffer[MAX_LINE + 1];
	VARSTRUCT fvar[MAX_VARIABLES];

 /*
  * File variables and counters
  */

    FILE        *msgfile;      /* conversion program messages */
    FILE        *datafile;     /* input data file             */
    FILE		*stdfile;      /* standards file              */
    FILE		*deffile;      /* defaults file               */
    FILE		*segfile;      /* segment names file          */
    FILE		*namefile;     /* default segment names file  */
    int         inCounter=0;   /* input line counter          */
    int         outCounter=0;  /* output line counter         */
	int			linesSkipped=0; /* lines other than a DDE     */

/**************************************************************/
/*     Allocate and initialize variable areas                 */
/**************************************************************/

	memset(segName, 0, sizeof(segName));
	memset(defaultnames, 0, sizeof(defaultnames));
	memset(currType, 0, sizeof(currType));
	memset(fvar, 0, MAX_VARIABLES * sizeof(VARSTRUCT));
	for (i=0; i < 50; i++)
	{
		lastAtLevel[i] = 0;
	}

/**************************************************************/
/*     Load the repeating names file into memory.             */
/**************************************************************/

	loadRepeat();

/**************************************************************/
/*  Read the default segment names file into memory.          */
/**************************************************************/

	namefile = fopen("defnames.dat", "rb");
	if (namefile != NULL)
	{
		fseek(namefile, 0L, SEEK_END);
		length = ftell(namefile);
		fseek(namefile, 0L, SEEK_SET);
		fread(tempArea, 1, length, namefile);
		fclose(namefile);

		// change any new lines into end of string characters
		i = 0;
		j = 0;
		while(j < length)
		{
			if (tempArea[j] != '\n')
			{
				if (tempArea[j] != '\r')
				{
					defaultnames[i] = tempArea[j];
				}

				i++;
			}

			j++;
		}
	}

/**************************************************************/
/* The existing standards file, if any, will be read into     */
/* memory.  This will allow for checking for duplicate        */
/* entries when running the utility against many data         */
/* dictionary files.                                          */
/**************************************************************/

/**************************************************************/
/* Open the existing standards file for reading.              */
/**************************************************************/

	stdfile = fopen("STDFILE.MTE", "rb");

	if (stdfile != NULL)
	{
		// determine the length of the definitions data and read
		// it into memory
		fseek(stdfile, 0L, SEEK_END);
		length = ftell(stdfile);
		fseek(stdfile, 0L, SEEK_SET);
		currStdfile = (char *) malloc(length + 11);
		memset(currStdfile, 0, length + 11);
		fread(currStdfile, 1, length, stdfile);
		currStdfile[length] = 0;

		/* close the file */
		fclose(stdfile);
	}
	else
	{
		// indicate no standards data is available
		currStdfile = (char *) malloc(11);
		memset(currStdfile, 0, 11);
		length = 0;
	}

/**************************************************************/
/*     Open the standards file.                               */
/**************************************************************/

	stdfile = fopen("STDFILE.MTE", "wb");

	if (stdfile == NULL)
	{
		printf("Unable to open standards file\n");
		exit(10);
	}
	else
	{
		// start by writing the existing data
		if (currStdfile[0] > 0)
		{
			// Start by rewriting the existing data
			fwrite(currStdfile, 1, length, stdfile);
			fflush(stdfile);
		}
	}

	deffile = fopen("DEFFILE.MTE", "wb");

	if (deffile == NULL)
	{
		fclose(stdfile);
		printf("Unable to open defaults file\n");
		exit(11);
	}

	segfile = fopen("SEGNAME.MTE", "wb");

	if (segfile == NULL)
	{
		fclose(stdfile);
		fclose(deffile);
		printf("Unable to open segment name file\n");
		exit(12);
	}

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

	if (argc >= 2)        /* check for file names */
	{
		/* Get the file name parameter, making sure not to overflow */
		strncpy(dataFileName,argv[1],sizeof(dataFileName));

		/* check for help request */
		if ((dataFileName[0] == '?') || (dataFileName[1] == '?'))
		{
			printHelp(argv[0]);
			exit(1);
		} /* endif */
	}
	else
	{
		/* must have two file names, one for copy book and one for data file */
		printHelp(argv[0]);
		exit(1);
	} /* endif */


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

	/* get the full file name */
	strcpy(msgFileName,dataFileName);

	/* Remove any extension from the message file name */
	/* look for a period */
	fileExt = strchr(msgFileName, '.');
	if (fileExt != NULL)
	{
		/* remove the file extension if it exists */
		(*fileExt) = 0;
	} /* endif */

   /* add a file extension to the message files */
   strcat(msgFileName,".msg");

/**************************************************************/
/* Open the copy book, data file and message file.            */
/**************************************************************/

	if ((datafile = fopen(dataFileName, "r")) == NULL)
	{
		printf("Unable to open input file - %s\n", dataFileName);
		exit(100);
	}

	if ((msgfile = fopen(msgFileName, "a+")) == NULL)
	{
		fclose(datafile);
		printf("Unable to open message file - %s\n", msgFileName);
		exit(101);
	}

	/* Write the name of the input file to the message file */
	writeFileName(dataFileName, msgfile);

	/* skip over the "0" entries */
	memset(buffer, 0, sizeof(buffer));
	while ((buffer[20] == '0') || (buffer[20] == 0))
	{
		if (fgets(buffer, MAX_LINE_LEN, datafile) == NULL)
		{
			break;
		}

		/* Get rid of any new line characters at the end of the line */
		if (buffer[strlen(buffer) - 1] == '\n')
		{
			buffer[strlen(buffer) - 1] = 0;
		}
	}

/**************************************************************/
/*                                                            */
/*  This routine will read the ACORD ASC file                 */
/*  and generate a listing file and a meta data file.         */
/*                                                            */
/**************************************************************/

   /* Read the next line(s) in the source file. */
	while (buffer[3] == 'E')
	{
		/* clear the data area and move the */
		/* first entry into the data area   */
		memset(dataArea, 0, sizeof(dataArea));

		/* point to the data area */
		dde_ptr = (DDE_REC *) dataArea;

		/* capture the current group identifier */
		memcpy(currType, buffer + 20, 4);

		/* read in all the entries for a single group */
		while ((memcmp(currType, buffer + 20, 4) == 0) && (buffer[3] != 'T'))
		{
			/* capture this entry */
			memcpy(dde_ptr, buffer, sizeof(DDE_REC));

			/* move on to the next area in the buffer */
			dde_ptr++;

			/* Get the next line */
			memset(buffer, 0, sizeof(buffer));
			if (fgets(buffer, sizeof(buffer), datafile) == NULL)
			{
				memset(buffer, 0, sizeof(buffer));
				/* break if end of file is reached */
				break;
			}

			/* Get rid of any new line characters at the end of the line */
			if (buffer[strlen(buffer) - 1] == '\n')
			{
				buffer[strlen(buffer) - 1] = 0;
			}

			/* Count the lines read from the data dictionary file */
			inCounter++;
		}

		/* point to the first line in memory */
		dde_ptr = (DDE_REC *) dataArea;

		/* clear the current name data areas */
		memset(nameArea, 0, sizeof(nameArea));
		nameLen=0;
		memset(UnameArea, 0, sizeof(UnameArea));
		UnameLen=0;

		/* build the new data file name */
		memcpy(segName, dde_ptr->group_desig, sizeof(dde_ptr->group_desig));

		/* create a standards file entry and a defaults file entry */
		defdone = 0;
		memset(segVersion, 0, sizeof(segVersion));
		if ((segName[0] != '0') && (dde_ptr->ver_no[0] > '0'))
		{
			buildDefaultEntry(stdfile, msgfile, deffile, dde_ptr, segName, segVersion, currStdfile);
			defdone = 1;
		}

		/* reset the offset variable */
		curoffset = 0;

		/* create a single group element at the 01 level */
		memset(fvar, 0, MAX_VARIABLES * sizeof(VARSTRUCT));
		buildSegNameEntry(segfile, msgfile, dde_ptr, segName, tempName);
		fvar[0].VarLevel = 1;
		fvar[0].VarChild = 1;
		fvar[0].VarLastChild = 1;
		fvar[0].VarOccur = 1;
		level = 5;
		maxvar = 1;
		lastAtLevel[level] = 1;

		/* Remember this is the current parent */
		saveParent = 0;

		/* get the name of this segment */
		memset(tempName, 0, sizeof(tempName));
		memcpy(tempName, dde_ptr->group_desig, sizeof(dde_ptr->group_desig));

		/* strip the trailing blanks */
		while ((strlen(tempName) > 0) && (tempName[strlen(tempName) - 1] == ' '))
		{
			tempName[strlen(tempName) - 1] = 0;
		}

		/* indicate the segment we are processing */
		fprintf(msgfile, "Processing %s (%s)\n", tempName, segName);
		fflush(msgfile);

		/* process the entries we just read into memory */
		while (dde_ptr->reference_name[0] != 0)
		{
			/* process next element */
			/* get the length of this item */
			memcpy(workArea, dde_ptr->element_length, sizeof(dde_ptr->element_length));
			workArea[sizeof(dde_ptr->element_length)] = 0;
			length = atoi(workArea);

			/* update the last child pointer for the root record */
			fvar[0].VarLastChild = maxvar - 1;

			/* check for a header record */
			if (memcmp(dde_ptr->element_name, "Header", 6) == 0)
			{
				/* special case the header */
				maxvar = buildHeader(fvar, length);

				/* increment the output counter */
				outCounter += maxvar;

				/* the header is the last entry at level 5 for now */
				lastAtLevel[level] = 1;

				/* Root element is the parent now */
				saveParent = 0;

				/* initialize the current offset to the end of the header */
				curoffset = length;

				/* go on to the next entry */
				dde_ptr++;
			}
			else
			{
				// Not a header element
				if (defdone == 0)
				{
					buildDefaultEntry(stdfile, msgfile, deffile, dde_ptr, segName, segVersion, currStdfile);
					defdone = 1;
				}

				/* Ignore any variables that are zero length */
				if (length > 0)
				{
					/* check for a repeating data item */
					repPtr = checkRepeat(dde_ptr, &repType, &occurs);

					if ((occurs > 1) && (repType == 'C'))
					{
						/* This item is a repeating group item */
						/* first, add the group item as a parent */
						buildHeaderEntry(&fvar[maxvar],
										 level,
										 curoffset,
										 0,
										 saveParent,
										 maxvar + 1,
										 maxvar + 1,
										 0,
										 lastAtLevel[level]);
						
						/* correct the number of occurs */
						fvar[maxvar].VarOccur = occurs;

						/* update the chain pointers */
						fvar[lastAtLevel[level]].VarNextLevel = maxvar;
						lastAtLevel[level] = maxvar;

						/* add a name to the name table */
						if (repPtr->newname[0] > 0)
						{
							saveName(repPtr->newname);
						}
						else
						{
							/* create a valid variable name of less than 30 characters */
							memset(tempName, 0, sizeof(tempName));

							/* create a name for this data element */
							buildName(dde_ptr, tempName, maxvar);

							/* remember the variable name */
							saveName(tempName);
						}

						saveParent = maxvar;
						maxvar++;

						/* increase the level by 5 */
						level += 5;

						/* get count of number of data elements in this group */
						j = repPtr->last;

						/* figure out how many variables to skip */
						skipCount = (occurs - 1) * j;

						while (j > 0)
						{
							memcpy(workArea, dde_ptr->element_length, sizeof(dde_ptr->element_length));
							workArea[sizeof(dde_ptr->element_length)] = 0;
							currlen = atoi(workArea);

							/* create additional entries */
							buildHeaderEntry(&fvar[maxvar],
											 level,
											 curoffset,
											 currlen,
											 saveParent,
											 0,
											 0,
											 0,
											 0);

							/* increment the lines processed counter */
							outCounter++;

							/* add the name to the name tables */
							memset(tempName, 0, sizeof(tempName));

							/* create a valid variable name of less than 30 characters */
							/* create a name for this data element */
							buildName(dde_ptr, tempName, maxvar);

							/* remember the variable name */
							saveName(tempName);

							/* Is this the only child? */
							if (fvar[saveParent].VarChild != maxvar)
							{
								/* not the first child */
								/* update the peer pointers */
								fvar[maxvar].VarPrevLevel = maxvar - 1;
								fvar[maxvar - 1].VarNextLevel = maxvar;
							}

							/* make this variable the last child of the parent */
							fvar[saveParent].VarLastChild = maxvar;

							/* move on to the next variable */
							maxvar++;

							/* move on to the next offset */
							curoffset += currlen;

							/* update the parent lengths */
							fvar[saveParent].VarLen   += currlen;
							fvar[0].VarLen += currlen * occurs;

							/* see if we have done all the variables */
							j--;

							/* increment the lines processed counter */
							outCounter++;

							/* go on to the next variable */
							dde_ptr++;
						}

						/* decrease the level by 5 */
						level -= 5;

						/* increment the offset by the extra occurs */
						curoffset += (occurs - 1) * fvar[saveParent].VarLen;

						/* restore the current parent */
						saveParent = fvar[saveParent].VarParent;

						/* skip over the variables we grouped together */
						dde_ptr += skipCount;
					}
					else
					{
						/* we can handle this item as a single item */
						/* set the level, parent, length and number of occurences */
						fvar[maxvar].VarLevel = level;
						fvar[maxvar].VarParent = saveParent;
						fvar[maxvar].VarLen = length;
						fvar[maxvar].VarOccur = occurs;

						/* get the offset */
						fvar[maxvar].VarOffset = curoffset;

						/* get the data type */ 
						saveType(dde_ptr, fvar, maxvar);

						// Update the previous NextAtLevel pointer for this level
						fvar[lastAtLevel[level]].VarNextLevel = maxvar;
						fvar[maxvar].VarPrevLevel = lastAtLevel[level];
						lastAtLevel[level] = maxvar;

						/* create a valid variable name of less than 30 characters */
						memset(tempName, 0, sizeof(tempName));

						/* check if we found this element in the REPEAT.DAT file */
						if (repPtr != NULL)
						{
							if (repPtr->newname[0] != 0)
							{
								strcpy(tempName, repPtr->newname);
							}
						}

						/* create a name for this data element if we didn't find one */
						if (tempName[0] == 0)
						{
							buildName(dde_ptr, tempName, maxvar);
						}

						/* remember the variable name */
						fvar[maxvar].VarNamePtr = nameLen;
						fvar[maxvar].VarUnamePtr = UnameLen;
						saveName(tempName);

						/* make this variable the last child and if */
						/* necessary, the first child               */
						fvar[saveParent].VarLastChild = maxvar;
						if (fvar[saveParent].VarChild == 0)
						{
							fvar[saveParent].VarChild = maxvar;
						}

						/* update the lengths of all the parents of this child */
						parent = saveParent;
						while (parent > 0)
						{
							fvar[parent].VarLen += length * occurs;
							parent = fvar[parent].VarParent;
						}

						/* update the root element, which is always a parent */
						fvar[0].VarLen += length * occurs;

						/* go on to the next variable */
						curoffset += length * occurs;
						maxvar++;
			
						/* go on to the next entry, skipping any recurring items */
						dde_ptr += occurs;

						/* increment the lines processed counter */
						outCounter++;
					} /* not a repeating group item */
				}	
				else
				{
					/* Length is zero - ignore this item */						
					dde_ptr++;
				}
			}	/* Not a header */
		}

		/* write out the variables into the mtd file */
		writeVars(fvar, segName, segVersion, maxvar, msgfile);

		/* tell what the next offset was */
		fprintf(msgfile, "Next offset was %d\n", curoffset);

		/* Write the number of lines read to the message file */
		writeLinesRead(outCounter, msgfile, segName);

	} /* endwhile */

	// write out a segment name to long name mapping file
	writeSegnameFile(segfile);

	/* write the counters to the message file */
	fprintf(msgfile,"\nTotals\nInput lines read in file %5d\n",inCounter);
	fprintf(msgfile, "\nTotal lines skipped %d\n", linesSkipped);

	/* close the files */
	fclose(datafile);
	fclose(stdfile);
	fclose(deffile);
	fclose(segfile);
	fclose(msgfile);

	/* Release the memory we have allocated */
	free(currStdfile);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

	return RC;
}

/*-------------*/
/* End Of File */
/*-------------*/
