//  metadata.cpp 

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <afxmt.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include <BipCpi.h>
#include "BipSampPluginUtil.h"

#include "mtdstruc.h"
#include "metamem.h"
#include "metadata.h"
#include "trace.h"
#include "parsubs.h"
#include "common.h"

#define ACORDMETADIR		"ACORDMETADIR"
#define METADATA_DIR_DEF    "C:\\acord"
#define STDFILE_NAME		"STDFILE.MTE"
#define DEFFILE_NAME		"DEFFILE.MTE"
#define SEGFILE_NAME		"SEGNAME.MTE"
#define MAX_TCG_ENTRIES		20
#define SEGNAMELEN			4
#define STDFILELEN			10
#define DEFFILELEN			6

#ifdef _DEBUG
static char Level[]="metadata.cpp V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="metadata.cpp V1.0 Release version ("__DATE__" "__TIME__")";
#endif

int     useCounter=0;               // Counter for least recently used meta data files
int     filesOpened=0;				// files added to cache
int     filesRemoved=0;				// files removed from cache
int     filesMax=0;                 // Maximum number of files in cache

FILETBL* fileTable=NULL;	        // File table to cache recent files 

static char * standards=NULL;				// Pointer to standards file contents
static char * defaults=NULL;				// Pointer to default file contents
static char * segname=NULL;					// Pointer to group names data
static char fileDirName[256];               // Location of metadata files (mtd) 
static CCriticalSection fileTableLock;		// semaphore for file maniplulation

#ifdef _DEBUG
extern int debug;
extern int traceMgmt;
extern int traceModule;
#endif

////////////////////////////////////
//
// Routine to find and read the
// contents of a file into a
// specified area.
//
////////////////////////////////////

int getFileData(const char * fileName, char * * dataArea)

{
	int		length=0;
	char	stdfilename[512];
	FILE	*stdfile;
#ifdef _DEBUG
	char	errmsg[512];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getFileData(%s)\n", fileName);
	}
#endif

	// get the drive and directory for the metadata files
	strcpy(stdfilename, fileDirName);
	strcat(stdfilename, fileName);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("Opening file %s\n", stdfilename);
	}
#endif

	// try and open the meta data standards file
	stdfile = fopen(stdfilename, "r");
	if (stdfile != NULL)
	{
		// determine the length of the data and read
		// it into memory
		fseek(stdfile, 0L, SEEK_END);
		length = ftell(stdfile);
		fseek(stdfile, 0L, SEEK_SET);
		(*dataArea) = (char *) malloc(length + STDFILELEN + 1);
		memset((*dataArea), 0, length + STDFILELEN + 1);
		fread((*dataArea), 1, length, stdfile);
		(*dataArea)[length] = 0;
		fclose(stdfile);
	}
	else
	{
#ifdef _DEBUG
		if (debug) 
		{
			sprintf(errmsg, "Unable to open file (%s)\n", stdfilename);
			Log(errmsg);
			traceError(__FILE__, __LINE__, errmsg);
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getFileData(length = %d)\n", length);
	}
#endif

	return length;
}

////////////////////////////////////
//
// Routine to find and read the
// contents of the meta data
// defaults file.
//
////////////////////////////////////

void getDefaultsData()

{
	int	length;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getDefaultsData()\n");
	}
#endif

	length = getFileData(DEFFILE_NAME, &defaults);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getDefaultsData(), length = %d\n", length);
	}
#endif
}

////////////////////////////////////
//
// Routine to find and read the
// contents of the meta data
// group names file.
//
////////////////////////////////////

void getSegNameData()

{
	int	length;
	int i;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getSegNameData()\n");
	}
#endif

	length = getFileData(SEGFILE_NAME, &segname);

	// now, we must turn any new line characters into zeros
	i = 0;
	while (i <= length)
	{
		// check for a new line character
		if (segname[i] == '\n')
		{
			// found a new line, make it the end of an entry
			segname[i] = 0;
		}

		// go on to the next character
		i++;
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getSegNameData(), length = %d\n", length);
	}
#endif
}

////////////////////////////////////
//
// Routine to find and read the
// contents of the meta data
// standards file.
//
////////////////////////////////////

void getStdfileData()

{
	int	length;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getStdfileData()\n");
	}
#endif

	length = getFileData(STDFILE_NAME, &standards);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getStdfileData(), length = %d\n", length);
	}
#endif
}

////////////////////////////////////////
//
// Routine to look up the standard that
// a particular segment is defined in.
//
// The 3 character standard number is
// moved into the version field that is
// passed in.  If the segment name is
// not found, the version will be set
// to nulls.
//
////////////////////////////////////////

void segToStd(const char * segName, char * version)

{
	char *ptr;

	// initialize the returned value
	version[0] = 0;

	// return if there is no standards file data
	if (standards == NULL)
	{
		return;
	}

	ptr = standards;
	while ((ptr[0] != 0) && (memcmp(ptr + 5, segName, SEGNAMELEN) != 0))
	{
		ptr += STDFILELEN;
	}

	if (ptr[0] != 0)
	{
		memcpy(version, ptr, 3);
	}
}

//////////////////////////////////////
//
// Routine to look up a default version
// in the defaults file data and 
// return the version character that
// should be appended to the segment
// name.
//
// If the segment is not found,
// then a null character will
// be returned.
//
//////////////////////////////////////

void getDefaultVersion(const char *segName, char * segVersion)

{
	char	searchName[10];
	char	*ptr;

	// Initialize the segVersion variable to zeros
	segVersion[0] = 0;
	segVersion[1] = 0;
	segVersion[2] = 0;

	// get the segment name and version
	memcpy(searchName, segName, SEGNAMELEN);
	searchName[SEGNAMELEN] = 0;

	// check if we found the defaults file
	if (defaults != NULL)
	{
		// We found it, we need to work our way through the data
		ptr = defaults;
		while ((ptr[0] != 0) && (memcmp(searchName, ptr, SEGNAMELEN) != 0))
		{
			ptr += DEFFILELEN;
		}

		// Did we find this group name
		if (ptr[0] != 0)
		{
			// Do not return a blank version number
			if (ptr[SEGNAMELEN] > ' ')
			{
				segVersion[0] = ptr[SEGNAMELEN];
				if (ptr[SEGNAMELEN + 1] > ' ')
				{
					segVersion[1] = ptr[SEGNAMELEN + 1];
				}
			}

#ifdef _DEBUG
			if (traceModule) 
			{
				Log("Default version for group %s found - version %s\n", segName, segVersion);
			}
#endif
		}
		else
		{
#ifdef _DEBUG
			if (traceModule) 
			{
				Log("Unable to find default version for %s\n", searchName);
			}
#endif
		}
	}
}

//////////////////////////////////////
//
// Routine to look up a file standard
// in the standards file data and 
// return the version character that
// should be appended to the segment
// name.
//
// If the standard or segment is not
// found, then a null character will
// be returned.
//
//////////////////////////////////////

char getStandardVersion(const char *segName, const char *stdno)

{
	char	ch=0;
	char	searchName[10];
	char	*ptr;

	// get the segment name and version
	memcpy(searchName, stdno, 5);
	memcpy(searchName + 5, segName, SEGNAMELEN);
	searchName[9] = 0;

	// check if we found the standards file
	if (standards != NULL)
	{
		// We found it, we need to work our way through the data
		ptr = standards;
		while ((ptr[0] != 0) && (memcmp(searchName, ptr, 9) != 0))
		{
			ptr += STDFILELEN;
		}

		if (ptr[0] != 0)
		{
			ch = ptr[9];

#ifdef _DEBUG
			if (traceModule) 
			{
				Log("Standards data for segment %s standard %5.5s group version %c\n", segName, ptr, ch);
			}
#endif
		}
		else
		{
#ifdef _DEBUG
			if (traceModule) 
			{
				Log("Unable to find standards data for %s\n", searchName);
			}
#endif
		}
	}

	return ch;
}

void initializeStandards()

{
	if (standards != NULL)
	{
		free(standards);
		standards = NULL;
	}

	if (defaults != NULL)
	{
		free(defaults);
		defaults = NULL;
	}

	if (segname != NULL)
	{
		free(segname);
		segname = NULL;
	}

	getStdfileData();
	getDefaultsData();
	getSegNameData();
}

//////////////////////////////////////
//
// Routine to perform meta data file
// initialization.  It will return a
// pointer to the meta data file
// table.
//
//////////////////////////////////////

void initializeMetadata()

{
	char*			fileVar;

	/* Initialize the file table for the meta data cache in memory */
	fileTable = (FILETBL *) malloc(sizeof(FILETBL) * MAX_FILES);
	if (fileTable != NULL)
	{
		memset(fileTable, 0, sizeof(FILETBL) * MAX_FILES);
	}

	// Initiatize the file table lock
	InitializeCriticalSection(fileTableLock);

	/* check for meta file directory name in environment */
	memset(fileDirName, 0, sizeof(fileDirName));
	fileVar = getenv(ACORDMETADIR);
	if (fileVar != NULL)
	{
		// Get the directory name for the meta data files
		// from the environmental variable MTDDIRECTORY
		strncpy(fileDirName, fileVar, sizeof(fileDirName) - 2);
	}
	else
	{
		// set a default
		strcpy(fileDirName, METADATA_DIR_DEF);
	}

	// check for a backslash on the end of the name
	if ((strlen(fileDirName) > 0) && (fileDirName[strlen(fileDirName) - 1] != '\\'))
	{
		// append a trailing back slash character
		fileDirName[strlen(fileDirName)] = '\\';
	}

	getStdfileData();
	getDefaultsData();
	getSegNameData();
}

/**************************************************************/
/*                                                            */
/* Routine to reset the counter areas used to determine the   */
/*  least recently used meta data file.                       */
/*                                                            */
/**************************************************************/

void resetLRU()

{
	int	i;

	for (i=0; i < MAX_FILES; i++)
	{
		fileTable[i].useCount = 0;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to remove an entry from the metadata file table.   */
/*                                                            */
/**************************************************************/

void removeEntry(const int fileno)

{
	// release the storage for the oldest file
	if (fileTable[fileno].fvar != NULL)
	{
		free(fileTable[fileno].fvar);
	}

	// clear the oldest entry in the table
	fileTable[fileno].useCount = 0;
	memset(fileTable[fileno].fileName, DELETED_FILE_BYTE, sizeof(fileTable[fileno].fileName));
	fileTable[fileno].flags = 0;
	fileTable[fileno].fvar = NULL;
	fileTable[fileno].maxvar = 0;
	fileTable[fileno].timesUsed = 0;

	// decrement the number of files counter
	filesMax--;
}

/**************************************************************/
/*                                                            */
/* Routine to find an entry in the table.  If no entry is     */
/*  found, then an entry will be released and an attempt      */
/*  made to load the file into memory.  If the file cannot    */
/*  be loaded, an index of -1 will be returned.               */
/*                                                            */
/**************************************************************/

int findEntry(const char * segName, const int traceSw)

{
	int	i=0;
	int minUse = 0;
	int minUseCount=0;
	char defFileName[255 + 32];	// name of copy book definitions file 
	META_FILE_HEADER hdr;	// metadata file header section
	char errmsg[512];
    FILE *deffile;			// COBOL definitions file
	char *ptr;				// temporary pointer
	int	defLen=0;			// length of definition data 

	minUseCount = fileTable[0].useCount + 1;

	// see if we already have this file in the cache
	while ((i < MAX_FILES) && (fileTable[i].fileName[0] > 0) &&
			(strcmp(fileTable[i].fileName, segName) != 0))
	{
		// keep track of the least used file
		if ((fileTable[i].useCount < minUseCount) && (fileTable[i].inUseCount == 0))
		{
			minUse = i;
			minUseCount = fileTable[i].useCount;
		}

		// Always pick the first deleted entry for reuse if one is found
		if ((fileTable[i].fileName[0] == DELETED_FILE_BYTE) && (fileTable[minUse].fileName[0] != DELETED_FILE_BYTE)) 
		{
			minUse = i;
			minUseCount = fileTable[i].useCount;
		}

		// increment the counter
		i++;
	}

#ifdef _DEBUG
	if (traceSw > 0)
	{
		Log("Meta data index entry is %d, minUse %d %d\n", i, minUse, minUseCount);
	}
#endif

	if ((i >= MAX_FILES) || (fileTable[i].fileName[0] == 0))
	{
		// check if we have to reuse an existing entry
		if (i >= MAX_FILES)
		{
			// clear the least recently used file entry
			removeEntry(minUse);

			// set i to point to the first entry
			i = minUse;

			// increment the files removed from cache counter
			filesRemoved++;
		}

		// try to read the file data
		// build the definitions file name based on the 
		// message segment identifier
		memset(defFileName, 0, sizeof(defFileName));

		// Get the drive and directory for the meta data files
		strcpy(defFileName, fileDirName);

		// make sure we have a delimiter between the path name and the file name
		if ((strlen(defFileName) > 0) && (defFileName[strlen(defFileName) - 1] != '\\'))
		{
			strcat(defFileName, "\\");
		}

		// append the segment name as the first part of the file name
		strcat(defFileName, segName);

		// Finish the file name by appending the meta data file type
		strcat(defFileName, ".mtd");

#ifdef _DEBUG
		if (traceSw)
		{
			Log("Opening meta data file - %s\n", defFileName);
		}
#endif

		// Try to open the definitions file
		if ((deffile = fopen(defFileName, "rb")) != NULL)
		{
			// determine the length of the definitions data and read
			// it into memory
			// get the length of the metadata
			fseek(deffile, 0L, SEEK_END);
			defLen = ftell(deffile) - META_HDR_LEN;

			// read the header section of the metadata file
			fseek(deffile, 0L, SEEK_SET);
			fread(&hdr, 1, META_HDR_LEN, deffile);

			// allocate memory for the metadata and read it
			fileTable[i].fvar = (VAR_FILE_REC *) malloc(defLen);
			memset(fileTable[i].fvar, 0, defLen);
			fread(fileTable[i].fvar, 1, defLen, deffile);

			// get the number of variable table entries from the header
			fileTable[i].maxvar = hdr.Maxvar;
			fileTable[i].dataLength = hdr.Seglen;

			// Calculate the pointers for the names and initialization values
			fileTable[i].namePtr = (char *) fileTable[i].fvar + hdr.Names - sizeof(hdr);
			ptr = (char *) fileTable[i].fvar + hdr.Unames - sizeof(hdr);
			fileTable[i].UnamePtr = (CciChar *) ptr;
			fileTable[i].initChar = hdr.InitChar;

			// check if we have an initialization string
			if (hdr.InitialValue > 0)
			{
				fileTable[i].initString = ptr + hdr.InitialValue - hdr.Unames;
			}

			fileTable[i].timesUsed = 1;

#ifdef _DEBUG
			if (traceSw > 0)
			{
				Log("Number of bytes read %d into 0x%x\n", defLen,fileTable[i].fvar);
			}
#endif

			// close the file
			fclose(deffile);

			// Increment the file counters
			filesOpened++;
			filesMax++;

			// set the segment name
			memset(fileTable[i].fileName, 0, sizeof(fileTable[i].fileName));
			strncpy(fileTable[i].fileName, segName, sizeof(fileTable[i].fileName) - 1);
		}
		else
		{
			sprintf(errmsg, "Unable to open meta data file - %s\n", defFileName);

#ifdef _DEBUG
			if (debug)
			{
				Log(errmsg);
			}
#endif

			// Indicate file not found
			i = -1;
		}
	}

	return i;
}

/**************************************************************/
/*                                                            */
/* Routine to remove a meta data cache element from memory.   */
/*  This is invoked with a special management message, to     */
/*  allow for online updating of a meta data file without     */
/*  having to shut down the broker.                           */
/*                                                            */
/**************************************************************/

void removeMetaDataEntry(const char *segName)

{
	int i;
	int found=0;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> removeMetaDataEntry() for %s\n", segName);
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
	fileTableLock.Lock();

	i = 0;
	while (i < MAX_FILES)
	{
		if (memcmp(fileTable[i].fileName, segName, strlen(segName)) == 0)
		{
			// found a matching entry, now try to delete it
			found = 1;

			if (fileTable[i].inUseCount == 0)
			{
				// not currently in use, we can delete it safely
				memset(fileTable[i].fileName, DELETED_FILE_BYTE, sizeof(fileTable[i].fileName));
				fileTable[i].useCount = 0;

#ifdef _DEBUG
				if (traceMgmt) 
				{
					Log("Segment(%s) removed from cache\n", segName);
				}
#endif
			}
			else
			{
				// The file is currently being used
				// flag it for later deletion
				fileTable[i].flags = fileTable[i].flags | DELAYED_DELETION;
				fileTable[i].useCount = 0;

#ifdef _DEBUG
				if (traceMgmt) 
				{
					Log("Segment(%s) scheduled for later removal\n", segName);
				}
#endif
			}
		}

		i++;
	}

#ifdef _DEBUG
	if (found == 0)
	{
		// the entry is not in the cache
		if (traceMgmt) 
		{
			Log("Segment(%s) not found in cache\n", segName);
		}
	}
#endif

	// release the semaphore
	fileTableLock.Unlock();

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- removeMetaDataEntry()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to remove all meta data cache elements from memory.*/
/*  This is invoked with a special management message, to     */
/*  allow for online updating of a meta data file without     */
/*  having to shut down the broker.                           */
/*                                                            */
/**************************************************************/

void removeAllMetaData()

{
	int i;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> removeAllMetaData()\n");
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
	fileTableLock.Lock();

	i = 0;
	while (i < MAX_FILES)
	{

		// found the entry, now try to delete it
		if (fileTable[i].inUseCount == 0)
		{
			// not currently in use, we can delete it safely
			memset(fileTable[i].fileName, DELETED_FILE_BYTE, sizeof(fileTable[i].fileName));
			fileTable[i].useCount = 0;
		}
		else
		{
			// The file is currently being used
			// flag it for later deletion
			fileTable[i].flags = fileTable[i].flags | DELAYED_DELETION;
			fileTable[i].useCount = 0;
		}

		i++;
	}

	// Finally, flush the standards and defaults files
	initializeStandards();

	// release the semaphore
	fileTableLock.Unlock();

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- removeAllMetaData()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to decrment the in use counter in a meta data      */
/*  cache element, and if need be, remove the entry from      */
/*  memory.                                                   */
/*                                                            */
/**************************************************************/

void decrementUseCount(int traceSw, const int fileIdx)

{
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> decrementUseCount() fileIdx(%d)\n", fileIdx);
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
	fileTableLock.Lock();

	// check if the inUseCounter is > 0
	if (fileTable[fileIdx].inUseCount > 0)
	{
		fileTable[fileIdx].inUseCount--;

		// check for a delayed deletion request
		if (((fileTable[fileIdx].flags & DELAYED_DELETION) > 0) && (fileTable[fileIdx].inUseCount == 0))
		{
			// Yes, we need to delete this entry now 
			fileTable[fileIdx].flags = 0;
			memset(fileTable[fileIdx].fileName, DELETED_FILE_BYTE, sizeof(fileTable[fileIdx].fileName) - 1);
			fileTable[fileIdx].useCount = 0;

#ifdef _DEBUG
			if (traceSw) 
			{
				Log("Cache file entry deleted for fileIdx(%d)\n", fileIdx);
			}
#endif
		}
	}
	else
	{
		// something is wrong - the counter should be > 0
#ifdef _DEBUG
		if (debug) 
		{
			Log("Invalid use count for fileIdx(%d) inUseCount(%d)\n",
				fileIdx, fileTable[fileIdx].inUseCount);
		}
#endif
	}
	
	// release the semaphore
	fileTableLock.Unlock();

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- decrementUseCount()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to get the meta data from the meta data file.      */
/*  It keeps a cache of meta data files in memory. The        */
/*  segment name is assumed to be 4 characters long and       */
/*  is contained in the input data.                           */
/*                                                            */
/**************************************************************/

int getMetaData(const int traceSw,
				const char *versionPtr,
				const char *segName,
				const CciChar group)

{
	int rc;
	int i;
	int	j;
	const char* ptr=NULL;
	char stdVer[3];
	char groupVer;
	char stdno[4];
	char segVersion[3];
	char fullName[7];
	int	RC=-1;				// default return code

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getMetaData() for (%s) versionPtr(0x%x) group (%C)\n", 
			segName, versionPtr, group);
	}
#endif

	// Search for the segment data in memory before we 
	// read the file.  First, we must build the full
	// reference name, which is the name of the meta data
	// file.  It consists of the four-character group
	// identifier with either the group version or the
	// standard version and modification level appended.
	groupVer = 0;
	memset(stdVer, 0, sizeof(stdVer));

	// Get the group identifier as the first part of the name
	memset(fullName, 0, sizeof(fullName));
	memcpy(fullName, segName, SEGNAMELEN);

	// Check if a group version number was contained in the segment
	// If so, we will use it.  If not, we will try to get one from
	// the standard version and revision.
	if (group > chDigit_0)
	{
		// The group version number was specified explicitly
		cciUcsToMbs(&rc, &group, &groupVer, 1, DEF_CODEPAGE);
	}
	else
	{
		// get the version number in the directory, if one was found.
		// This takes several steps.  First, the segment name
		// must be looked up in the standards file and the
		// standard number must be ascertained.  The standard
		// number is used to find the version and revision
		// entry for the standard in the 2TCG group segment.
		// Finally, the segment name, version and revision
		// information is used to find the segment version
		// number for the particular group segment.  The
		// characters are then appended to the group (segment)
		// identifier to locate the appropriate meta data
		// file.
		if ((versionPtr != NULL) && (standards != NULL))
		{
			// Initialize the standard number
			memset(stdno, 0, sizeof(stdno));

			// Next, see if we can find the standard that this segment is defined in
			segToStd(segName, stdno);

			// We now can search the version number table and see if we can
			// find a matching entry in the 2TCG segment
			if (stdno[0] != 0)
			{
				// We found the segment name and the standard
				// now, try to find the corresponding entry
				// in the 2TCG data area
				j = 0;
				ptr = versionPtr;
				while ((j < MAX_TCG_ENTRIES) && (memcmp(ptr, stdno, 3) != 0) &&
						(ptr[0] > ' '))
				{
					// move on to the next entry
					ptr += 5;
					j++;
				}

				if (memcmp(ptr, stdno, 3) == 0)
				{
					// capture the standard version and revision if they are
					// present and not equal to "00"
					if ((ptr[3] > ' ') && (memcmp(ptr + 3, "00", 2) != 0))
					{
						memcpy(stdVer, ptr + 3, 2);
					}

					groupVer = getStandardVersion(fullName, ptr);
				}
			}
		}
	}

	// Now, append the version qualifier, if any
	if (groupVer > '0')
	{
		fullName[SEGNAMELEN] = groupVer;
	}
	else
	{
		// See if we have a standard version and revision
		if (stdVer[0] > '0')
		{
			// Yes, we found a version and revision
			// Use it to determine the right meta data file
			strcat(fullName, stdVer);
		}
	}

	// get the semaphore, so that only one thread at a time is
	// working on the cache
	fileTableLock.Lock();

#ifdef _DEBUG
	if (traceSw)
	{
		Log("Getting meta data for segment %s\n", fullName);
	}
#endif

	// Either find this entry or clear a space for it in the table
	i = findEntry(fullName, traceSw);

	// check if we found the file
	// If not, we want to see if we can find the entry in the
	// defaults file
	if (i < 0)
	{
		// Didn't find default either - Log the error
#ifdef _DEBUG
		if (debug)
		{
			Log("Could not find metadata for %s\n", fullName);
		}
#endif

		// We didn't find the entry, try for a default entry
		// Get the group identifier as the first part of the name
		memset(fullName, 0, sizeof(fullName));
		memcpy(fullName, segName, SEGNAMELEN);

		// Try the default file with no version number
		getDefaultVersion(segName, segVersion);
		strcat(fullName, segVersion);

		// Either find this entry or clear a space for it in the table
		i = findEntry(fullName, traceSw);
	}

	// check if we found the file yet
	// If not, as a final try, use just the segment name 
	// for a default file name
	if (i < 0)
	{
		// Didn't find default either - Log the error
#ifdef _DEBUG
		if (debug)
		{
			Log("Could not find metadata for %s\n", fullName);
		}
#endif

		// We didn't find the entry, try for a system default entry
		// Get the group identifier as the first part of the name
		memset(fullName, 0, sizeof(fullName));
		memcpy(fullName, segName, SEGNAMELEN);

		// Either find this entry or clear a space for it in the table
		i = findEntry(fullName, traceSw);
	}

	// Did we find any version?
	if (i < 0)
	{
		// Didn't find default either - Log the error
#ifdef _DEBUG
		if (debug)
		{
			Log("Could not find metadata for %s\n", fullName);
		}
#endif

		RC = -1;
	}
	else
	{
		// increment the usage counter
		useCounter++;
		if (useCounter > (INT_MAX - 5))
		{
			resetLRU();
			useCounter = 1; 
		}

		fileTable[i].useCount = useCounter;
		fileTable[i].timesUsed++;
		fileTable[i].inUseCount++;

		// pass back an index to the data in memory
		RC = i;
	}

	fileTableLock.Unlock();

#ifdef _DEBUG
	if (traceModule > 0)
	{
		Log("PLUGIN: <- getMetaData() for (%s) fileIdx(%d)\n", 
			segName, RC);
	}
#endif

	return RC;
}

/**************************************************************/
/*                                                            */
/* Routine to take a long group name and return the four      */
/*  character group name that it corresponds to.  If the      */
/*  long name is not recognized, then the segName parameter   */
/*  will contain a null string.                               */
/*                                                            */
/**************************************************************/

void getGroupName(char * segName, const char * longName)

{
	char *ptr;

	// set a default return value
	segName[0] = 0;

	// check if we have a segment name file
	if (segname != NULL)
	{
		// get the first entry in the table
		ptr = segname;

		// search for this one
		while ((ptr[0] != 0) && (strcmp(ptr + SEGNAMELEN, longName) != 0))
		{
			ptr += strlen(ptr) + 1;
		}

		// found it - return the group identifier as a string
		if (ptr[0] != 0)
		{
			memcpy(segName, ptr, SEGNAMELEN);
			segName[SEGNAMELEN] = 0;
		}
	}
}

/**************************************************************/
/*                                                            */
/* Routine to take a long group name and return the four      */
/*  character group name that it corresponds to.  If the      */
/*  long name is not recognized, then the segName parameter   */
/*  will contain a null string.                               */
/*                                                            */
/**************************************************************/

void getLongName(const char * segName, char * longName, int maxchar)

{
	char *ptr;

	// set a default return value
	longName[0] = 0;

	// check if we have a segment name file
	if (segname != NULL)
	{
		// get the first entry in the table
		ptr = segname;

		// search for this one
		while ((ptr[0] != 0) && (memcmp(ptr, segName, SEGNAMELEN) != 0))
		{
			ptr += strlen(ptr) + 1;
		}

		// found it - return the long name as a string
		if (ptr[0] != 0)
		{
			strncpy(longName, ptr + SEGNAMELEN, maxchar - 1);
		}
	}
}

const char * getMetadataLevel()

{
	return Level;
}