//  parsubs.cpp 

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <afxmt.h>
#include <wchar.h>
#include <windows.h>
#endif

#include <stdio.h>
#include <BipCci.h>
#include <BipCpi.h>
#include <BipCni.h>
#include "parsubs.h"
#include "mtdstruc.h"
#include "BipSampPluginUtil.h"
#include "trace.h"
#include "common.h"

// Parser exception codes (built by message compiler)
#include "acorderr.h"

#ifdef WIN32
#define CHARSETL			L"CodedCharSetId"
#define ENCODINGL			L"Encoding"
#endif

#ifdef _DEBUG
	extern int	traceModule;
#endif

#ifdef _DEBUG
static char Level[]="parsubs.cpp V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="parsubs.cpp V1.0 Release version ("__DATE__" "__TIME__")";
#endif

///////////////////////////////////////////////////////////
//
// ErrorToString
//
///////////////////////////////////////////////////////////

char* ErrorToString(int iRc)

{
	static char	message[256];
	switch(iRc)
	{
	case CCI_INV_FACTORY_NAME:
		return("Invalid factory name");
	case CCI_INV_FACTORY_OBJECT:        
		return("Invalid factory object");
	case CCI_INV_NODE_NAME:      
		return("Invalid node name");
	case CCI_INV_NODE_OBJECT:          
		return("Invalid node object");
	case CCI_INV_TERMINAL_NAME:        
		return("Invalid terminal name");
	case CCI_INV_TERMINAL_OBJECT:      
		return("Invalid terminal object");
	case CCI_INV_MESSAGE_OBJECT:       
		return("Invalid message object");
	case CCI_INV_ELEMENT_OBJECT:       
		return("Invalid element object");
	case CCI_INV_PARSER_NAME:          
		return("Invalid parser name");
	case CCI_INV_PARSER_OBJECT:        
		return("Invalid parser object");
	case CCI_INV_VFTP:                 
		return("Invalid VFTP");
	case CCI_INV_DATA_POINTER:
		return("Invalid data pointer");
	case CCI_INV_LENGTH:               
		return("Invalid length");
	case CCI_NAME_EXISTS:              
		return("Name already exists");
	case CCI_BUFFER_TOO_SMALL:
		return("Buffer too small");
	case CCI_NO_BUFFER_EXISTS:         
		return("Missing or invalid buffer");
	case CCI_PARSER_NAME_TOO_LONG:
		return("Parse name too long");
	case CCI_NO_EXCEPTION_EXISTS:    
		return("No exception exists");
	case CCI_INV_MESSAGE_CONTEXT:     
		return("Invalid message context");
	case CCI_INV_DATA_BUFLEN:     
		return("Invalid data buffer length");
	case CCI_MISSING_IMPL_FUNCTION:
		return("Missing IMPL function");
	case CCI_INV_OBJECT_NAME:   
		return("Invalid object name");
	case CCI_INV_LOG_TYPE:         
		return("Invalid log type");
	case CCI_INV_SQL_EXPR_OBJECT:
		return("Invalid SQL expression object");
	case CCI_INV_TRANSACTION_TYPE:     
		return("Invalid transaction type");
	case CCI_INV_STATEMENT:    
		return("Invalid statement");
	case CCI_EXCEPTION:           
		return("Exception");
	case CCI_EXCEPTION_UNKNOWN:
		return("Unknown Exception");
	case CCI_EXCEPTION_FATAL:       
		return("Fatal Exception");
	case CCI_EXCEPTION_RECOVERABLE:
		return("Recoverable Exception");
	case CCI_EXCEPTION_CONFIGURATION:
		return("Configuration Exception");
	case CCI_EXCEPTION_PARSER: 
		return("Parser Exception");
	case CCI_EXCEPTION_CONVERSION:
		return("Conversion Exception");
	case CCI_EXCEPTION_USER:    
		return("User Exception");
	default:
		sprintf(message, "Error: Unknown code: %d", iRc);
		return(message);
	}
}

///////////////////////////////////////////////////////////
//
// Log errors to MQSI V2 Broker trace file
//
///////////////////////////////////////////////////////////

void traceError(char * srcfile, const int srcline, char *szFormat, ...)

{
	int	rc;
	va_list	list;
	va_start(list, szFormat);
	char	templine[4192];

	vsprintf(templine, szFormat, list);

	// write the error to the MQSI V2 trace file
	cciLog(&rc, CCI_LOG_ERROR, srcfile, srcline, "traceError", 
			ERRDICT, CCILOG_MSG, templine, templine, 0);

	va_end(list);
}

/**************************************************************/
/*                                                            */
/* Routine to find the value element associated with a        */
/*  particular element and return it.  If the input element   */
/*  has a value associated with it, then that element is      */
/*  used.  If a name element is passed in, then the first     */
/*  child will be used, if it exists and is a value element.  */
/*                                                            */
/**************************************************************/

CciElement * getValueElement(CciElement* nameElement)

{
	int	RC;
	CciElementType	elementtype=CCI_ELEMENT_TYPE_UNKNOWN;
	CciElement* valueElement=CCI_NULL_ADDR;

	elementtype = cpiElementType(&RC, nameElement);
	if (RC == 0)
	{
		// check what type of element has been passed in
		// if it is a name element, then we will see if there is a
		// value element as the first child.  If this is the case,
		// we will use that.  If the element passed in has a value
		// we will extract that value directly
		if (elementtype == CCI_ELEMENT_TYPE_NAME)
		{
			valueElement = cpiFirstChild(&RC, nameElement);

			// check if this is a value element
			elementtype = cpiElementType(&RC, valueElement);

			// make sure it is a value type
			// ignore the RC, since elementtype is previously
			// set to CCI_ELEMENT_TYPE_NAME
			if (elementtype != CCI_ELEMENT_TYPE_VALUE)
			{
				// not a value type, reset the pointer to null
				valueElement=CCI_NULL_ADDR;
			}
		}
		else
		{
			if ((elementtype == CCI_ELEMENT_TYPE_VALUE) || 
				(elementtype == CCI_ELEMENT_TYPE_NAME_VALUE))
			{
				valueElement = nameElement;
			}
		}
	}

	return valueElement;
}

/**************************************************************/
/*                                                            */
/* Routine to find the value of a particular element and      */
/*  return it as a character value.                           */
/*                                                            */
/**************************************************************/

int	findCharacterValue( char *value, int maxSize, CciElement* nameElement, int codepage)

{
	CciChar	Uvalue[65537];
	char	work[128];
	int		size;
	int		rc;
	int		len;
	int		result=1;
	CciInt	tempInt=0;
	CciValueType	valuetype=CCI_VALUE_TYPE_UNKNOWN;
	CciElement* valueElement=CCI_NULL_ADDR;

#ifdef _DEBUG
	// make sure we have somewhere to put the data
	ASSERT(maxSize > 0);
	ASSERT(value != NULL);
#endif

	// initialize the value to nulls
	value[0] = 0;

	valueElement = getValueElement(nameElement);

	// check if we have a value element
	if (valueElement != CCI_NULL_ADDR)
	{
		// found the value element - now try to get the value
		valuetype = cniElementValueType(&rc, valueElement);

		// we can ignore the return code, since valuetype is
		// initialized to CCI_VALUE_TYPE_UNKNOWN
		switch (valuetype)
		{
		case CCI_VALUE_TYPE_INTEGER:
			{
				tempInt = cpiElementIntegerValue(&rc, valueElement);
				if (rc == 0)
				{
					// convert the integer to a character value
					_i64toa(tempInt, work, 10);

					// get the length of the data
					len = strlen(work);

					// check if it is too long
					if (len >= maxSize)
					{
						len = maxSize - 1;
					}

					// move the data to the output area
					memcpy(value, work, len);
					value[len] = 0;

					// indicate success
					result = 0;
				}

				break;
			}
		case CCI_VALUE_TYPE_CHAR:
			{
				size = cpiElementCharacterValue(&rc, valueElement, Uvalue, sizeof(Uvalue) - 1);
				if (rc == CCI_SUCCESS)
				{
					len = size;	

					// check if it is too long
					if (len >= maxSize)
					{
						len = maxSize - 1;
					}

					// convert the data back from Unicode
					size = cciUcsToMbs(&rc, Uvalue, value, len, codepage); 
					value[len] = 0;

					// indicate success
					result = rc;
				}

				break;
			}
		}
	}

	return result;
}

/***************************************************************/
/*                                                             */
/* Routine to find the value element for a particular          */
/*  variable. The input variable is given as a fully qualified */
/*  name (e.g. a.b.c.d), but without the first element name.   */
/*                                                             */
/***************************************************************/

CciElement * findElementByName(const CciChar * fullName, CciElement * firstElement)

{
	int			rc;
	int			size;
	CciElement*	child;
	CciChar		name[512];
	CciChar		nameSeg[512];
	CciChar*	nextName;
	CciChar*	currName;

	// get a copy of the fully qualified name
	CciCharCpy(nameSeg, fullName);
	nextName = nameSeg;
	currName = nameSeg;

	// find the first child of the element passed
	child = cpiFirstChild(&rc, firstElement);

	while ((rc == CCI_SUCCESS) && (child != CCI_NULL_ADDR) && (nextName != NULL))
	{
		// isolate this name segment
		nextName = wcschr(currName, '.');
		if (nextName != NULL)
		{
			nextName[0] = 0;

			// point to the next level
			nextName++;
		}

		// Now, get the name of the first element
		size = cpiElementName(&rc, child, name, sizeof(name) - 1);

		// we have the name, now see if it is what we are looking for
		while ((rc == CCI_SUCCESS) && (CciCharCmp(name, currName) != 0) && (child != CCI_NULL_ADDR))
		{
			// Not this one, try the next one
			child = cpiNextSibling(&rc, child);
			size = cpiElementName(&rc, child, name, sizeof(name) - 1);
		}

		if ((rc == CCI_SUCCESS) && (child != CCI_NULL_ADDR) && (nextName != NULL))
		{
			// move to the next level
			child = cpiFirstChild(&rc, child);
			currName = nextName;
		}
	}

	return child;
}

/**************************************************************/
/*                                                            */
/* Routine to search all the children of a particular name    */
/*  element for a particular named element.                   */
/*                                                            */
/**************************************************************/

CciElement * findNextElement(CciElement* element, CciChar * varName) 

{
	int			rc;
	int			skipCount=0;
	CciSize		size;
	CciChar		Uname[MAX_NAME_LEN + 1];

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(Uname, 0, sizeof(Uname));
		if (element != CCI_NULL_ADDR)
		{
			size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1);
		}

		Log("PLUGIN: -> findNextElement() element %S (0x%x)\n", Uname, element);
	}
#endif

	while (element != CCI_NULL_ADDR)
	{
		size = cpiElementName(&rc, element, Uname, sizeof(Uname));
		if (rc == CCI_SUCCESS)
		{
			// check if this is the name we are looking for
			if (CciCharCmp(Uname, varName) == 0)
			{
				break;
			}

			element = cniNextSibling(&rc, element);
			skipCount++;
		}
		else
		{
			// unable to get the name of the element
			// return a null element address
			element = CCI_NULL_ADDR;
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(Uname, 0, sizeof(Uname));
		if (element != CCI_NULL_ADDR)
		{
			size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1);
		}

		Log("PLUGIN: <- findNextElement() element %S (0x%x) skipped(%d)\n", 
			Uname, element, skipCount);
	}
#endif

	return element;
}

/**************************************************************/
/*                                                            */
/* Routine to find the root element of the properties.  It    */
/* should be the first child of the real root.                */
/*                                                            */
/* Note - use the node functions to ensure parsing of the     */
/* elements in case they are not yet parsed.                  */
/*                                                            */
/**************************************************************/

CciElement * getPropertiesElement(CciParser* parser, CciChar* propertyName)

{
	int	rc;
	CciElement*	root;
	CciElement* properties;
	CciElement* child=CCI_NULL_ADDR;

	// first, get the parser root element
	root = cpiRootElement(&rc, parser);
	if (rc == CCI_SUCCESS)
	{
		// get the real root element
		root = cpiParent(&rc, root);

		// did we get the real root?
		if (rc == CCI_SUCCESS)
		{
			// got the real root - now get the first child,
			// which should be the Properties element
			properties = cniFirstChild(&rc, root);

			// did we get the properties element?
			if ((rc == CCI_SUCCESS) && (properties != CCI_NULL_ADDR))
			{
				// got the properties - now get the first child,
				// search for the name we want
				properties = cniFirstChild(&rc, properties);

				// did we get the first child of the properties element?
				if ((rc == CCI_SUCCESS) && (properties != CCI_NULL_ADDR))
				{
					// search for the element we want
					child = findNextElement(properties, propertyName);
				}
			}

		}
	}

	return child;
}

/**************************************************************/
/*                                                            */
/* Routine to find the code page in the properties of a       */
/*  message and return the value.                             */
/*                                                            */
/**************************************************************/

int	getCodePage(CciParser* parser)

{
	int	codepage=0;
	int	rc;
	char value[256];
	CciElement* element;

	element = getPropertiesElement(parser, CHARSETL);

	// check if we found the element
	if (element != CCI_NULL_ADDR)
	{
		value[0] = 0;
		rc = findCharacterValue(value, sizeof(value) - 1, element, DEF_CODEPAGE);
		codepage = atoi((char *) value);
	}

	return codepage;
}

/**************************************************************/
/*                                                            */
/* Routine to find the numeric encoding in the properties of  */
/*  a message and return the value.                           */
/*                                                            */
/**************************************************************/

int	getEncoding(CciParser* parser)

{
	int	encoding=0;
	int	rc;
	char	value[256];
	CciElement* element;

	element = getPropertiesElement(parser, ENCODINGL);

	// check if we found the element
	if (element != CCI_NULL_ADDR)
	{
		value[0] = 0;
		rc = findCharacterValue(value, sizeof(value) - 1, element, DEF_CODEPAGE);
		encoding = atoi((char *) value);
	}

	return encoding;
}

const char * getParsubsLevel()

{
	return Level;
}