/**************************************************************/
/*                                                            */
/* Program Name:  PRINTMTD.C                                  */
/*                                                            */
/*  Program to print the contents of an mtd data file.        */
/*                                                            */
/*  Written by Jim MacNair - April, 2000                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  Current Version 1.00                                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#include "mtdstruc.h"


static char copyright[] = "(C) Copyright IBM Corp, 2000";
static char Version[]=\
"@(#)printmtd V1.00 - program to print mtd files  - Jim MacNair ("__DATE__")";

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

int printReport(VAR_FILE_REC *fvar,
				char * nameArea,
				const int maxvar, 
				FILE *msgfile)

{
	int rc=0;
	int i;
	char usage[6];
	char* namePtr;
	char* dependPtr;

	// Print a heading
	fprintf(msgfile,"\n    Lev   Ofs  Leng   Type  Occ  Max Parent  Children   Siblings   Flags Name                              Depending\n");
	fprintf(msgfile,"                                             1ST  Last  Next  Prev  1  2\n\n");

	// Print the individual lines
	for (i=0; i < maxvar; i++)
	{
		if (fvar[i].OccurDepend > 0)
		{
			dependPtr = nameArea + fvar[fvar[i].OccurDepend].Nameptr;
		}
		else
		{
			dependPtr = "";
		}

		switch (fvar[i].Usage)
		{
		case USE_DISPLAY:
			{
				strcpy(usage, "Char");
				break;
			}
		case USE_BINARY:
			{
				strcpy(usage, "Int");
				break;
			}
		case USE_POINTER:
			{
				strcpy(usage, "Ptr");
				break;
			}
		case USE_PACKED:
			{
				strcpy(usage, "Pack");
				break;
			}
		case USE_FLOAT1:
		case USE_FLOAT2:
			{
				strcpy(usage, "Float");
				break;
			}
		case USE_INDEX:
			{
				strcpy(usage, "Index");
				break;
			}
		default:
			{
				usage[0] = 0;
				break;
			}
		}

		// point to the name in the name area
		namePtr = nameArea + fvar[i].Nameptr;

		fprintf(msgfile, "%4d %2d %5d %5d %6s %4d %4d %5d %5d %5d %5d %5d %2x %2x %-32s %s\n",
				i,
				fvar[i].Level,
				fvar[i].Offset,
				fvar[i].Length,
				usage,
				fvar[i].Occur,
				fvar[i].OccurMax,
				fvar[i].Parent,
				fvar[i].FirstChild,
				fvar[i].LastChild,
				fvar[i].NextSibling,
				fvar[i].PrevSibling,
				fvar[i].Flags,
				fvar[i].Flags2,
				namePtr,
				dependPtr);
	}

	return rc;
}

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void writeFileName(char *fileName, FILE *msgfile)

{
	/* write the file name to the message file */
	fprintf(msgfile, "Processing file %s\n", fileName);
}

/**************************************************************/
/*                                                            */
/* Routine to display the command syntax.                     */
/*                                                            */
/**************************************************************/

void printHelp(char *cmd)
{
   printf("\nPRINTMTD V1.00\n");
   printf("\nCommand usage is:\n\n");
   printf("  %s mtd-file-name\n\n", cmd);
   printf("A report file is printed with with the name of\n");
   printf("the mtd file and an extension of msg\n");
}

/**************************************************************/
/*                                                            */
/*                                                            */
/*                                                            */
/**************************************************************/

int main(int argc, char *argv[])

{
	int  maxvar=0;			// Number of fields found            
	int	dataLen=0;			// length of COBOL data               
	VAR_FILE_REC * fvar;	// Pointer to metadata file variable definitions 
	META_FILE_HEADER* hdr;	// Pointer to metadata file header area
	char *nameArea;			// Pointer to metadata file name area
	char *fileExt;			// file extension                     
	char msgFileName[255];	// name of message file               
	char mtdFileName[255];	// name of mtd file                   
	int	RC=0;				// Return code from program           

 /*
  * File variables and counters
  */

    FILE        *mtdfile;	// input COBOL program         
    FILE        *msgfile;	// conversion program messages 

/**************************************************************/
/*     Allocate and initialize variable areas                 */
/**************************************************************/

	memset(mtdFileName, 0, sizeof(mtdFileName));
	memset(msgFileName, 0, sizeof(msgFileName));

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

	if (argc > 1)        // check for file name 
	{
		// Get the file name parameter, making sure not to overflow 
		strncpy(mtdFileName,argv[1],sizeof(mtdFileName) - 1);

		// check for help request 
		if ((mtdFileName[0] == '?') || (mtdFileName[1] == '?'))
		{
			printHelp(argv[0]);
			exit(1);
		} // endif 
	}
	else
	{
		// must have file name for mtd file 
		printHelp(argv[0]);
		exit(1);
	} // endif 


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

	// get the full file name 
	strcpy(msgFileName,mtdFileName);

	// Remove any extension from the message file name 
	// look for a period 
	fileExt = strchr(msgFileName, '.');
	if (fileExt != NULL)
	{
		// remove the file extension if it exists 
		(*fileExt) = 0;
	} // endif 

	// add a file extension to the message files 
	strcat(msgFileName,".msg");

/**************************************************************/
/* Open the mtd file and message file.                        */
/**************************************************************/

	if ((mtdfile = fopen(mtdFileName, "rb")) == NULL)
	{
		printf("Unable to open input file - %s\n", mtdFileName);
		exit(100);
	}

	if ((msgfile = fopen(msgFileName, "w")) == NULL)
	{
		fclose(mtdfile);
		printf("Unable to open message file - %s\n", msgFileName);
		exit(101);
	}

	// Write the name of the input file to the message file 
	writeFileName(mtdFileName, msgfile);

	// determine the length of the data file and read it into memory
	fseek(mtdfile, 0L, SEEK_END);
	dataLen = ftell(mtdfile);
	fseek(mtdfile, 0L, SEEK_SET);
	nameArea = malloc(dataLen);
	fread(nameArea, 1, dataLen, mtdfile);

	// Now, calculate the pointers to different parts of the file
	hdr = (META_FILE_HEADER *) nameArea;
	nameArea += META_HDR_LEN;
	fvar = (VAR_FILE_REC *) nameArea;
	nameArea += (hdr->Maxvar) * sizeof(fvar[0]);

	// calculate the number of variable table entries
	maxvar = hdr->Maxvar;

	RC = printReport(fvar, nameArea, maxvar, msgfile);

	// print the header fields
	fprintf(msgfile, "\nHeader Length %d\n", hdr->HdrLength);
	fprintf(msgfile, "Version       %d\n", hdr->Version);
	fprintf(msgfile, "Group length  %d\n", hdr->Seglen);
	fprintf(msgfile, "Max vars      %d\n", hdr->Maxvar);
	fprintf(msgfile, "Names offset  %d\n", hdr->Names);
	fprintf(msgfile, "Unicode ofs   %d\n", hdr->Unames);
	fprintf(msgfile, "Init String   %d\n", hdr->InitialValue);

	/* close the files */
	fclose(mtdfile);
	fclose(msgfile);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

	return RC;
}

/*-------------*/
/* End Of File */
/*-------------*/
